import platform
import os

def get_config_dir():
  if platform.system() == "Windows":
    return os.path.join(os.environ["LOCALAPPDATA"], "mailcalaid")
  return os.path.expanduser("~/.config/mailcalaid")



import argparse
import os
import logging
import mailbox
from datetime import datetime
from mailcalaid.mail import Pop3Client, ImapClient

def mailboxes_command(args):
  for mailbox in args.client.list_mailboxes():
    print(mailbox)

def list_command(args):
  msg_id = (args.page - 1) * args.page_size + 1
  msg_id_end = msg_id + args.page_size - 1
  for msg in args.client.fetch_messages(msg_id, msg_id_end, headeronly=not args.full):
    print("{0:3} {1} {2:40} {3}".format(msg.msg_id, msg.date.isoformat() if msg.date else "?", msg.sender[:38], msg.subject))

def download_command(args):
  download = mailbox.mbox(args.download, create=True)
  # download.lock()
  if not args.id:
    args.id = 1
  if not args.id_end:
    args.id_end = args.client.total_messages
  for msg in args.client.fetch_messages(args.id, args.id_end):
    download.add(msg.message)
  download.flush()
  # archive.unlock()

def show_command(args):
  msg_id = args.id
  if args.id < 0:
    msg_id = args.client.total_messages + args.id + 1
  msg = args.client.fetch_message(msg_id)
  print("-------------------------------------------------")
  print("Subject  ", msg.subject)
  print("Date     ", msg.date)
  print("From     ", msg.sender)
  print()
  print(msg.plain or msg.html)

def delete_command(args):
  if args.id:
    args.client.mark_deleted(args.id)
  elif args.keep:
    args.client.mark_deleted_keep(args.keep)
  elif args.before:
    args.client.mark_deleted_before(args.before)
  elif args.after:
    args.client.mark_deleted_after(args.after)
  elif args.all:
    args.client.mark_deleted_all()
  args.client.flush()

parser = argparse.ArgumentParser()
parser.add_argument("--proto", default=os.getenv("MAIL_PROTO"), help="pop3, imap")
parser.add_argument("--host", default=os.getenv("MAIL_HOST"), help="mail server host")
parser.add_argument("--port", default=os.getenv("MAIL_PORT"), type=int, help="mail server port")
parser.add_argument("--user", default=os.getenv("MAIL_USER"), help="mail user")
parser.add_argument("--passwd", default=os.getenv("MAIL_PASSWD"), help="mail password")
parser.add_argument("--ssl", default=os.getenv("MAIL_SSL"), action="store_true", help="use ssl")
parser.add_argument("--debug", action="store_true", help="show debugging log")
parser.add_argument("--dry-run", action="store_true", help="swallow all writing/deleting operations")
parser.add_argument("--mailbox", default="INBOX", help="select remote mailbox (imap only)")
# parser.add_argument("--batch", type=int, default=100, help="batch size when process massive amount of records. e.g. fetching thousands of messages.")

subparsers = parser.add_subparsers(title='subcommands',
                                   description='valid subcommands',
                                   help='additional help')

parser_mailboxes = subparsers.add_parser("mailboxes", help="list mailboxes (imap only)")
parser_mailboxes.set_defaults(command=mailboxes_command)

parser_list = subparsers.add_parser("list", help="list messages")
parser_list.add_argument("page", nargs='?', type=int, default=1, help="page number")
parser_list.add_argument("-s", "--page-size", type=int, default=10, help="page size")
parser_list.add_argument("-f", "--full", action="store_true", help="fetch full message instead of header only")
parser_list.set_defaults(command=list_command)

parser_show = subparsers.add_parser("show", help="show message")
parser_show.add_argument("id", type=int, help="message id")
parser_show.set_defaults(command=show_command)

parser_download = subparsers.add_parser("download", help="download messages")
parser_download.add_argument("download", help="download mbox file")
parser_download.add_argument("--id", type=int, help="message id / start id")
parser_download.add_argument("--id-end", type=int, help="message id end")
parser_download.set_defaults(command=download_command)

parser_delete = subparsers.add_parser("delete", help="delete messages")
parser_delete.add_argument("id", nargs='?', help="message id to be deleted")
parser_delete.add_argument("--all", action="store_true")
parser_delete.add_argument("--keep", type=int)
parser_delete.add_argument("--after", type=lambda s: datetime.strptime(s, '%Y-%m-%d').astimezone())
parser_delete.add_argument("--before", type=lambda s: datetime.strptime(s, '%Y-%m-%d').astimezone())
parser_delete.set_defaults(command=delete_command)

def help_command(args):
  if args.name:
    subparsers.choices[args.name].print_help()
  else:
    parser.print_help()
parser_help = subparsers.add_parser("help", help="show help")
parser_help.add_argument("name", nargs='?', help="command name")
parser_help.set_defaults(command=help_command)

args = parser.parse_args()

logging.basicConfig(format='[%(asctime)s] %(name)s: %(message)s', level=logging.DEBUG if args.debug else logging.INFO)

if not (args.host and args.user and args.passwd):
  parser.print_usage()
  exit(1)

kwargs= {
  "host": args.host,
  "port": args.port,
  "user": args.user,
  "password": args.passwd,
  "ssl": args.ssl,
  "batch_size": args.batch,
  "dry_run": args.dry_run,
}
if args.proto == "pop3":
  args.client = Pop3Client(**kwargs)
elif args.proto == "imap":
  args.client = ImapClient(**kwargs)
  if args.mailbox:
    total_messages = args.client.select(args.mailbox)

args.command(args)


import json
import re
from time import sleep
import logging
import os
from urllib import request
from string import Template
from datetime import datetime, time, timedelta
from mailcalaid.mail import ImapClient, Pop3Client
from configparser import ConfigParser

logging.basicConfig(format='[%(asctime)s] %(name)s: %(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)

dry_run = False

config_datetime_fmt="%Y-%m-%d %H:%M:%S"
config_dir = os.getenv("CONFIG_DIR", "config" )

config_file = os.path.join(config_dir, "mail2bot.ini")
config = ConfigParser(interpolation=None)
config.read(config_file, encoding="utf-8")

state_file = os.path.join(config_dir, "mail2bot_state.ini")
state = ConfigParser(interpolation=None)
if os.path.exists(state_file):
  state.read(state_file, encoding="utf-8")
else:
  state.add_section("state")
  state["state"]["previous_started_at"] = ""

general_config = config["general"]
interval = general_config.getint("interval", 60)
workhours_start = general_config.getint("workhours_start", 9)
workhours_end = general_config.getint("workhours_end", 18)
cache_dir = general_config.get("cahce_dir", "cache")
if not cache_dir.startswith("/") and config_dir:
  cache_dir = os.path.join(config_dir, cache_dir)

server_config = config["imap"]
proto=server_config["proto"]
host=server_config["host"]
port=int(server_config["port"])
user=server_config["user"]
passwd=server_config["passwd"]
ssl=server_config.getboolean("ssl", True)
timeout=server_config.getboolean("timeout", 60)

filter_config = config["filter"]
subject_keyword = filter_config.get("subject_keyword")
fromaddrs = filter_config.get("fromaddrs")
if fromaddrs:
  fromaddrs = set(fromaddrs.split(","))
ignore_realnames = filter_config.get("ignore_realnames")
if ignore_realnames:
  ignore_realnames = set(filter(lambda x:x,  ignore_realnames.split("\n")))

bothook_config = config["bothook"]
link_re = re.compile(bothook_config["link_re"], re.M)
link_idx = bothook_config.getint("link_idx", fallback=0)
bothook_url=bothook_config["bothook_url"]
bothook_headers = {}
bothook_headers_config = config["bothook request headers"]
if bothook_headers_config:
  for header in bothook_headers_config:
    bothook_headers[header] = bothook_headers_config[header] 
bothook_body_tpl = bothook_config["bothook_body"]
bothook_body_tpl =  Template(bothook_body_tpl)

def notify_bothook(detail):
  subject = detail.subject
  localdate = datetime.fromtimestamp(detail.date.timestamp())
  links = link_re.findall(detail.text)
  if not links:
    logger.error(f"failed to extract the link:\n{detail.text}")
    return
  realname, fromaddr = detail.sender_addr
  body = (bothook_body_tpl.substitute(
    subject=subject,
    subject_json=json.dumps(subject),
    date=localdate,
    link=links[link_idx],
    realname=realname,
    fromaddr=fromaddr,
  ))
  req = request.Request(
    url=bothook_url,
    method="POST",
    data=body.encode("utf-8"),
  )
  for header in bothook_headers:
    req.add_header(header, bothook_headers[header])
  if dry_run:
    # logger.info(f"POST {bothook_url} with body:\n{body}")
    logger.info("notify for %s %s", localdate, subject)
    return
  with request.urlopen(req) as res:
    logger.info(f"notify for {subject} status: {res.status}")


def checkmail(previous_started_at: datetime) -> datetime:
  kwargs = dict(
    host=host,
    port=port,
    user=user,
    password=passwd,
    ssl=ssl,
    timeout=timeout,
  )
  client = ImapClient(**kwargs) if proto=="imap" else Pop3Client(**kwargs)
  for msg in client.fetch_messages_after(previous_started_at, headeronly=True):
    if subject_keyword not in msg.subject:
      continue
    realname, fromaddr = msg.sender_addr
    if fromaddr not in fromaddrs:
      continue
    if realname in ignore_realnames:
      continue
    notify_bothook(client.fetch_message(msg.msg_id))
  client.close()

state_config = state["state"]
def stateful_checkmail():
  previous_started_at = state_config.get("previous_started_at")
  if previous_started_at :
    previous_started_at = datetime.strptime(previous_started_at, config_datetime_fmt)
  if not previous_started_at:
    previous_started_at = datetime.today()-timedelta(days=1)
  if not previous_started_at.tzinfo:
    previous_started_at = previous_started_at.replace(tzinfo=datetime.now().astimezone().tzinfo)
  logger.info("start checking new mails since %s", previous_started_at)
  started_at = datetime.now()
  try:
    checkmail(previous_started_at)
    state_config["previous_started_at"] = started_at.strftime(config_datetime_fmt)
    if not dry_run:
      with open(state_file, "w", encoding="utf8") as f:
        state.write(f)
    logger.info("done checking new mails, next since would be %s", started_at)
  except Exception:
    logger.exception("failed to check new mails")



import argparse
parser = argparse.ArgumentParser()
parser.add_argument("--dry-run", action="store_true", default=False)
args = parser.parse_args()
dry_run = args.dry_run

from mailcalaid.cal.holiday import ChinaHolidayBook
cn_holiday_book=None
if not dry_run:
  cn_holiday_book = ChinaHolidayBook(
    cache_dir=cache_dir,
    workhours_start=time(hour=workhours_start),
    workhours_end=time(hour=workhours_end),
  )
while True:
  if dry_run or cn_holiday_book.is_workhour():
    stateful_checkmail()
  sleep(interval)


import logging
import imaplib
import re
from datetime import datetime
from typing import Generator, Union, List
from mailcalaid.mail.mailclient import MailClient, Message

logger = logging.getLogger(__name__)

# Ref https://www.rfc-editor.org/rfc/rfc3501#section-6.4.5

class ImapClient(MailClient):
  MSG_HEADER = '(BODY.PEEK[HEADER])'
  MSG_FULL = '(RFC822)'
  client: imaplib.IMAP4
  mailbox: str = "INBOX"

  def open(self):
    if self.ssl:
      self.client = imaplib.IMAP4_SSL(host=self.host, port=self.port)
    else:
      self.client = imaplib.IMAP4(host=self.host, port=self.port)
    self.client.socket().settimeout(self.timeout)
    code, resp = self.client.login(self.user, self.password)
    if code != 'OK':
      raise Exception(resp[0].decode())
    self.select(self.mailbox)

  def close(self):
    self.client.close()

  @property
  def total_messages(self) -> int:
    return self.select(self.mailbox)

  def list_mailboxes(self) -> Generator[dict, None, None]:
    """List mailboxes in the current account"""
    list_response_pattern = re.compile(
      r'\((?P<flags>.*?)\) "(?P<delimiter>.*)" (?P<name>.*)'
    )
    code, resp = self.client.list()
    if code != 'OK':
        raise Exception(resp[0].decode())
    for line in resp:
      mailbox = list_response_pattern.match(line.decode()).groupdict()
      yield mailbox

  def select(self, mailbox: str) -> int:
    """Select a mailbox and return the number of messages in it"""
    code, resp = self.client.select('"%s"' % mailbox)
    if code != 'OK':
      raise Exception(resp[0].decode())
    self.mailbox = mailbox
    return int(resp[0].decode())
  
  def _fetch_message(self, msg_id:int, headeronly: bool) -> bytes:
    message_parts = self.MSG_HEADER if headeronly else self.MSG_FULL
    code, resp = self.client.fetch(str(msg_id), message_parts)
    if code != 'OK':
        raise Exception(resp[0].decode())
    logger.debug("fetch messags %s, response length: %d", msg_id, len(resp)) 
    return resp[0][1]

  def search(self, criterion: str):
    """Search messages in the current mailbox"""
    code, resp = self.client.search(None, criterion)
    if code != 'OK':
      raise Exception(resp[0].decode())
    return resp[0].decode().split() if resp[0] else None

  def _mark_deleted(self, msg_id: int):
    code, resp = self.client.store(msg_id, "+FLAGS", "(\\Deleted)")
    if code != 'OK':
      raise Exception(resp[0].decode())

  def _unmark_deleted(self, msg_id: int):
    code, resp = self.client.store(msg_id, "-FLAGS", "(\\Deleted)")
    if code != 'OK':
      raise Exception(resp[0].decode())

  def _flush(self):
    code, resp = self.client.expunge()
    if code != 'OK':
      raise Exception(resp[0].decode())



import poplib
import email.utils
import email.header
import logging
from typing import Generator
from mailcalaid.mail.mailclient import MailClient, Message

logger = logging.getLogger(__name__)

class Pop3Client(MailClient):
  client: poplib.POP3

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def open(self):
    if self.ssl:
      pop3client = poplib.POP3_SSL(host=self.host, port=self.port, timeout=self.timeout)
    else:
      pop3client = poplib.POP3(host=self.host, port=self.port, timeout=self.timeout)
    pop3client.user(self.user)
    pop3client.pass_(self.password)
    self.client = pop3client

  def close(self):
    self.client.quit()

  @property
  def total_messages(self) -> int:
    return self.client.stat()[0]

  def total_size(self) -> int:
    """Total size of all messages in mailbox"""
    return self.client.stat()[1]

  def _fetch_message(self, msg_id:int, headeronly: bool) -> bytes:
    code, lines, octets = self.client.top(msg_id, 0) if headeronly else self.client.retr(msg_id)
    logger.debug("fetch message %d response code %s, octets %d", msg_id, code, octets)
    return b'\r\n'.join(lines)

  def _mark_deleted(self, msg_id: int):
    return self.client.dele(msg_id)

  def _flush(self):
    self.close()
    self.open()

  def unmark_all_deleted(self):
    """Unmark all deleted messages"""
    self.client.rset()



"""
Holiday
"""
from typing import Dict, Tuple, Generator
from datetime import date, datetime, timezone, time,timedelta, tzinfo
from abc import ABC, abstractmethod, abstractproperty
from functools import cached_property
from urllib import request
import json
import logging
import os
from mailcalaid.common import get_config_dir

logger = logging.getLogger(__name__)

class HolidayBook(ABC):
  """ Holiday Book for a country

  :param str cache_dir: cache directory for caching holiday data
  :param int workhours_start: start hour of work hours
  :param int workhours_end: end hour of work hours
  """
  holidays: Dict[date, Tuple[bool, str]]
  cache_dir: str
  workhours_start: time
  workhours_end: time

  def __init__(
    self,
    cache_dir:str="",
    workhours_start:time=time(hour=9),
    workhours_end:time=time(hour=18),
  ):
    if not cache_dir:
      cache_dir = os.path.join(get_config_dir(), "holiday")
    self.holidays = dict()
    self.cache_dir = cache_dir
    os.makedirs(cache_dir, exist_ok=True)
    s = self.sanitize_filename(self.country)
    if not s:
      raise Exception("country name is empty")
    self.cache_file = os.path.join(cache_dir, f"{s}.json")
    if os.path.exists(self.cache_file):
      try:
        self.load()
      except:
        pass
    self.workhours_start = workhours_start
    self.workhours_end = workhours_end

  @staticmethod
  def sanitize_filename(filename: str) -> str:
    return "".join(c for c in filename if c.isalnum() or c in (" ", ".", "_", "-"))

  @abstractmethod
  def load_year(self, year: int) -> Generator[Tuple[date, bool, str], None, None]:
    """Load holidays from some APIs for a given year
    """
    pass

  @abstractproperty
  def timezone(self) -> tzinfo:
    """Timezone
    """
    pass

  @abstractproperty
  def country(self) -> str:
    """Country code"""
    pass

  def mark(self, d: date, is_holiday: bool, name: str):
    """Mark a date as holiday or not"""
    self.holidays[d] = (is_holiday, name)

  def check(self, dt: date|datetime=None) -> Tuple[bool, str]:
    """Check if a date is holiday or not"""
    d = self.normalize_date(dt)
    self.ensure_year(d.year)
    mark = self.holidays.get(d.date())
    if mark:
      return mark
    is_weekend = d.weekday() in (5, 6)
    name = "weekend" if is_weekend else ""
    return is_weekend, name

  def is_holiday(self, dt: date|datetime=None) -> bool:
    is_holiday, _ = self.check(dt)
    return is_holiday

  def is_workhour(self, dt: date|datetime=None, extend:timedelta=None) -> bool:
    """Check if a datetime is work hour or not"""
    dt = self.normalize_date(dt)
    is_holiday, _ = self.check(dt)
    start = datetime.combine(dt.date(), self.workhours_start, tzinfo=self.timezone)
    end = datetime.combine(dt.date(), self.workhours_end, tzinfo=self.timezone) 
    if extend:
      start -= extend
      end += extend
    return not is_holiday and start <= dt < end

  def normalize_date(self, dt: date | datetime) -> datetime:
    """Normalize a date or datetime to a datetime with timezone
    timezone would be converted if given datetime has a different timezone to this book"""
    if isinstance(dt, datetime):
      dt = dt.astimezone(self.timezone)
    elif isinstance(dt, date):
      dt = datetime(dt.year, dt.month, dt.day, tzinfo=self.timezone)
    if dt is None:
      dt = datetime.now().astimezone()
    return dt

  def ensure_year(self, year):
    """Ensure holiday data for a given year is loaded, if not, load it from API"""
    for d in self.holidays.keys():
      if d.year == year:
        return
    logger.info("loading holiday data for %d", year)
    for d, is_holiday, name in self.load_year(year):
      self.mark(d, is_holiday, name)
    self.save()

  def next_workday(self, dt: date|datetime=None) -> datetime:
    """Get next workday"""
    dt = self.normalize_date(dt)
    while True:
      dt += timedelta(days=1)
      if not self.is_holiday(dt):
        return dt.date()

  def latest_workday(self, dt: date|datetime=None) -> datetime:
    """Get latest workday"""
    dt = self.normalize_date(dt)
    while True:
      if not self.is_holiday(dt):
        return dt.date()
      dt -= timedelta(days=1)

  def save(self):
    """Save holiday data to cache file"""
    holidays = dict()
    for d, mark in self.holidays.items():
      holidays[d.isoformat()] = mark
    with open(self.cache_file, "w", encoding="utf8") as f:
      json.dump(holidays, f, indent=2, ensure_ascii=False)

  def load(self):
    """Load holiday data from cache file"""
    with open(self.cache_file, "r", encoding="utf8") as f:
      holidays = json.load(f)
      for d, mark in holidays.items():
        self.holidays[date.fromisoformat(d)] = tuple(mark)


class ChinaHolidayBook(HolidayBook):
  """China holiday book based on Timor API"""

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  @cached_property
  def timezone(self) -> tzinfo:
    return timezone(timedelta(hours=8), "CST")

  @property
  def country(self) -> str:
    return "China"

  def load_year(self, year: int) -> Generator[Tuple[date, bool, str], None, None]:
    headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:77.0) Gecko/20100101 Firefox/77.0'}
    req = request.Request(f"https://timor.tech/api/holiday/year/{year}", headers=headers)

    with request.urlopen(req) as res:
      data = json.load(res)
      for item in data["holiday"].values():
        yield date.fromisoformat(item["date"]), item["holiday"], item["name"]


class NagerDateHolidayBook(HolidayBook):
  """Based on NagareDate API
  Supported Countries: https://date.nager.at/Country

  .. code-block:: text
    us_holiday_book = NagerDateHolidayBook(timedelta(hours=-7), "US", "cache")
    usd = datetime(2023, 1, 16, 0, 0, 0, 0, us_holiday_book.timezone)
    cnd = datetime(2023, 1, 16, 0, 0, 0, 0, cn_holiday_book.timezone)
    print("usa ", usd, ":  is_holiday, name = ", us_holiday_book.check(usd))
    # usa  2023-01-16 00:00:00-07:00 :  is_holiday, name =  (True, 'Martin Luther King, Jr. Day')
    print("usa ", cnd, ":  is_holiday, name = ", us_holiday_book.check(cnd))
    # usa  2023-01-16 00:00:00+08:00 :  is_holiday, name =  (True, 'weekend')

  :param utc_offset: UTC offset
  :param country_code: country code, see https://date.nager.at/Country
  """

  utc_offset: timedelta
  country_code: str

  def __init__(self, utc_offset: timedelta, country_code: str, *args, **kwargs):
    self.utc_offset = utc_offset
    self.country_code = country_code
    super().__init__(*args, **kwargs)

  @cached_property
  def timezone(self):
    return timezone(self.utc_offset, self.country_code)

  @property
  def country(self):
    return self.country_code

  def load_year(self, year: int) -> Generator[Tuple[date, bool, str], None, None]:
    headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:77.0) Gecko/20100101 Firefox/77.0'}
    req = request.Request(f"https://date.nager.at/api/v3/publicholidays/{year}/{self.country_code}", headers=headers)

    with request.urlopen(req) as res:
      data = json.load(res)
      for item in data:
        yield date.fromisoformat(item["date"]), True, item["name"]



if __name__ == "__main__":
  logging.basicConfig(format='[%(asctime)s] %(name)s: %(message)s', level=logging.INFO)
  cn_holiday_book = ChinaHolidayBook()
  us_holiday_book = NagerDateHolidayBook(timedelta(hours=-7), "US")
  print("cn_holiday_book tz:", cn_holiday_book.timezone.utcoffset(datetime.now()))
  print("us_holiday_book tz:", us_holiday_book.timezone.utcoffset(datetime.now()))
  print("china today:  is_holiday, name = ", cn_holiday_book.check())
  print("china now:  is_workhour = ", cn_holiday_book.is_workhour())
  print("usa today:  is_holiday, name = ", us_holiday_book.check())
  usd = datetime(2023, 1, 16, 0, 0, 0, 0, us_holiday_book.timezone)
  cnd = datetime(2023, 1, 16, 0, 0, 0, 0, cn_holiday_book.timezone)
  print("usa ", usd, ":  is_holiday, name = ", us_holiday_book.check(usd))
  print("usa ", cnd, ":  is_holiday, name = ", us_holiday_book.check(cnd))
  print("usa now:  is_workhour = ", us_holiday_book.is_workhour())
  usd = datetime(2023, 4, 7, 0, 0, 0, 0, us_holiday_book.timezone)
  cnd = datetime(2023, 4, 7, 0, 0, 0, 0, cn_holiday_book.timezone)
  print("usa ", usd, ":  is_holiday, name = ", us_holiday_book.check(usd))
  print("usa ", cnd, ":  is_holiday, name = ", us_holiday_book.check(cnd))
  cnd = datetime(2023, 4, 7, 6, 0, 0, 0, cn_holiday_book.timezone)
  print("usa ", cnd, ":  is_workhours(extend=2) = ", cn_holiday_book.is_workhour(extend=timedelta(hours=2)))
  cnd = date(2023, 4, 4)
  print("latest workday of", cnd, "is", cn_holiday_book.latest_workday(cnd))
  cnd = date(2023, 4, 5)
  print("latest workday of", cnd, "is", cn_holiday_book.latest_workday(cnd))
  print("next workday of", cnd, "is", cn_holiday_book.next_workday(cnd))


from dataclasses import dataclass
from datetime import datetime, timezone
from functools import cached_property
from typing import Tuple, Generator, Callable, List, Union, Optional
from abc import ABC, abstractmethod, abstractproperty
import email.utils
import email.header
import email.message
import logging

logger = logging.getLogger(__name__)

@dataclass
class Message:
  """Message wraps email.message.Message and provides some useful properties

  :param str msg_id: message id
  :param bytes msg: message bytes
  """
  msg_id: str
  msg: bytes

  def first_by_type(self, content_type: str) -> str:
    if self.message.is_multipart():
      for part in self.message.walk():
        if part.get_content_type() == content_type and "attachment" not in part.get("Content-Disposition", ""):
          return part.get_payload(decode=True).decode()
    else:
      return self.message.get_payload(decode=True).decode()

  @cached_property
  def message(self) -> email.message.Message:
    return email.message_from_bytes(self.msg)

  @cached_property
  def plain(self) -> str:
    return self.first_by_type("text/plain")

  @cached_property
  def html(self) -> str:
    return self.first_by_type("text/html")

  @cached_property
  def text(self)  -> str:
    return self.plain or self.html

  @cached_property
  def sender_addr(self) -> Tuple[str, str]:
    return email.utils.parseaddr(self.sender)

  @cached_property
  def sender(self):
    return decode_header(self.message["From"])

  @cached_property
  def to(self):
    return decode_header(self.message["To"])

  @cached_property
  def cc(self):
    return decode_header(self.message["Cc"])

  @cached_property
  def bcc(self):
    return decode_header(self.message["Bcc"])

  @cached_property
  def subject(self):
    return decode_header(self.message["Subject"]).replace("\r\n", "")

  @cached_property
  def date(self):
    d = self.message["Date"]
    if not d:
      d = self.message["Received"].split(";")[1].strip()
    if not d:
      logger.warning("no date header found\n%s", self.msg)
      return None
    try:
      d = email.utils.parsedate_to_datetime(decode_header(d))
      if d.tzinfo is None:
        d = d.replace(tzinfo=timezone.utc)
      return d
    except Exception as e:
      logger.warning("failed to parse date %s: %s", d, e)
      return None


def decode_header(header):
  # print("header", type(header), header)
  if not header:
    return ""
  value, charset = email.header.decode_header(header)[0]
  if isinstance(value, bytes):
    if charset:
      try:
        value = value.decode(charset)
      except Exception as e:
        logger.warning("failed to decode header %s: %s", header, e)
        value = value.decode("utf-8", errors="replace")
    else:
      value = str(value)
  return value


class MailClient(ABC):
  """Abstract mail client

  :param str host: mail server host
  :param int port: mail server port
  :param str user: mail server user
  :param str password: mail server password
  :param bool ssl: use ssl when connecting to mail server
  :param int batch_size: batch size when processing messages
  :param bool dry_run: dry run mode
  """

  def __init__(self, host: str, port: int, user: str, password: str, ssl=True, batch_size=100, dry_run=False, timeout=60):
    self.host = host
    self.port = port
    self.user = user
    self.password = password
    self.ssl = ssl
    self.batch_size = batch_size
    self.dry_run = dry_run
    self.timeout = timeout
    if self._fetch_message is None and self._fetch_messages is None:
      raise Exception("either _fetch_messages or _fetch_message must be implemented")
    self.open()

  @abstractmethod
  def open(self):
    """Open connection to mail server"""
    pass

  @abstractmethod
  def close(self):
    """Close connection to mail server"""
    pass

  @abstractproperty
  def total_messages(self) -> int:
    """Total number of messages in mailbox"""
    pass

  @abstractmethod
  def _mark_deleted(self, msg_id: int):
    pass

  def mark_deleted(self, msg_id: int):
    """Mark message as deleted"""
    logger.info("mark message %s as deleted", msg_id)
    if not self.dry_run:
      self._mark_deleted(str(msg_id))

  @abstractmethod
  def _flush(self):
    pass

  def flush(self):
    """Flush deleted messages"""
    logger.info("flushing")
    if not self.dry_run:
      self._flush()

  @abstractmethod
  def _fetch_message(self, msg_id:int, headeronly: bool) -> bytes:
    pass

  def fetch_message(self, msg_id: int=1, headeronly=False):
    """Fetch message

    :param int msg_id: message id
    :headeronly bool: fetch only header
    """
    logger.debug("fetching message %s, headeronly: %s", msg_id, headeronly)
    return Message(msg_id, self._fetch_message(msg_id, headeronly=headeronly))

  def fetch_messages(self,
    msg_id: Union[int,  List[int]],
    msg_id_end: int,
    headeronly=False,
  ) -> Generator[Message, None, None]:
    """Fetch messages
    
    :param int|list msg_id: message id or list of message ids
    :param int msg_id_end: optional, end message id
    :param bool headeronly: fetch only header
    """
    logger.debug("fetching messages %s - %s headeronly: %s", msg_id, msg_id_end, headeronly)
    if isinstance(msg_id, list):
      for i in msg_id:
        yield Message(i, self._fetch_message(i, headeronly=headeronly))
    else:
      step = 1 if msg_id < msg_id_end else -1
      for i in range(msg_id, msg_id_end + step, step):
        yield Message(i, self._fetch_message(i, headeronly=headeronly))

  def fetch_messages_after(self, dt: datetime, headeronly=True) -> Generator[Message, None, None]:
    """Fetch messages after date
    
    :param datetime dt: date
    :param bool headeronly: fetch only header
    """
    for msg in self.fetch_messages(self.total_messages, 1, headeronly=headeronly):
      if msg.date < dt:
        logger.debug("stop fetching because message %s date %s < %s", msg.msg_id, msg.date, dt)
        return
      else:
        logger.debug("message %s date %s", msg.msg_id, msg.date)
      yield msg

  def fetch_messages_before(self, dt: datetime, headeronly=True) -> Generator[Message, None, None]:
    """Fetch messages before date
    
    :param datetime dt: date
    :param bool headeronly: fetch only header
    """
    for msg in self.fetch_messages(1, self.total_messages, headeronly=headeronly):
      if msg.date > dt:
        logger.debug("stop fetching because message %s date %s > %s", msg.msg_id, msg.date, dt)
        return
      else:
        logger.debug("message %s date %s", msg.msg_id, msg.date)
      yield msg
  
  def mark_deleted_before(self, dt: datetime):
    """Mark messages before date as deleted """
    for msg in self.fetch_messages_before(dt):
      self.mark_deleted(msg.msg_id)
  
  def mark_deleted_after(self, dt: datetime):
    """Mark messages after date as deleted"""
    for msg in self.fetch_messages_after(dt):
      self.mark_deleted(msg.msg_id)
  
  def mark_deleted_keep(self, keep:int):
    """Mark messages as deleted while keeping the last n messages"""
    def batch():
      total = self.total_messages
      msg_id_end = total - keep
      if msg_id_end > self.batch_size:
        msg_id_end = self.batch_size
      if msg_id_end < 1:
        return False
      logger.info(f"total {total}, deleting 1 to {msg_id_end} messages")
      for msg_id in range(msg_id_end, 0, -1):
        if msg_id % 10 == 0:
          logging.debug("mark deleted %s",  msg_id)
        self.mark_deleted(msg_id)
      self.flush()
      return True
    while batch():
      pass

  def mark_deleted_all(self):
    """Mark all messages as deleted"""
    for msg_id in range(1, self.total_messages + 1):
      self.mark_deleted(msg_id)

  def unmark_deleted(self, msg_id: int):
    """Unmark message as deleted"""
    self.unmark_deleted(msg_id)